# -*- coding: utf-8 -*-
"""
Created on Fri Sep 24 10:08:24 2021

@author: Jean
"""

# Imports
# Ignore Warnings 
import warnings
warnings.filterwarnings('ignore')

# Basic Imports 
import numpy as np
import pandas as pd
import time

# Plotting 
import matplotlib.pyplot as plt

# Preprocessing
from sklearn.model_selection import train_test_split, KFold, cross_val_score

# Metrics 
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score

# ML Models
from lightgbm import LGBMRegressor 
from xgboost import XGBRegressor 
from catboost import CatBoostRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn import svm

# Model Tuning 
from bayes_opt import BayesianOptimization
from hyperopt import fmin, tpe, hp, Trials
from hyperopt.fmin import generate_trials_to_calculate

# Feature Importance 
import shap
    
# Reading a CSV File
# 9015
df_NN = pd.read_csv("D:/temp/data/Melbourne_housing/Melbourne_housing_pre.csv",  encoding="utf-8")

X=df_NN[['Year','YearBuilt','Distance','Lattitude','Longtitude','Propertycount',
          'Landsize','BuildingArea', 'Rooms','Bathroom', 'Car','Type_h','Type_t','Type_u']]
y=df_NN['LogPrice']
train_X, valid_X, train_y, valid_y = train_test_split(X,y, test_size = .20, random_state=42)

train_X2 = train_X.copy()
valid_X2 = valid_X.copy()

# Data standardization
mean = train_X.mean(axis=0)
train_X -= mean
std = train_X.std(axis=0)
train_X /= std
valid_X -= mean
valid_X /= std


##% evaluateRegressor
# from sklearn.metrics import mean_squared_error, mean_absolute_error
def evaluateRegressor(true,predicted,message = "Test set"):
    MSE = mean_squared_error(true,predicted,squared = True)
    MAE = mean_absolute_error(true,predicted)
    RMSE = mean_squared_error(true,predicted,squared = False)
    LogRMSE = mean_squared_error(np.log(true),np.log(predicted),squared = False)
    R2 = r2_score(true,predicted)
    print(message)
    print("MSE:", MSE)
    print("MAE:", MAE)
    print("RMSE:", RMSE)
    print("LogRMSE:", LogRMSE)
    print("R2 :", R2)
    
##% Plot True vs predicted values. Useful for continuous y 
def PlotPrediction(true,predicted, title = "Dataset: "):
    fig = plt.figure(figsize=(20,20))
    ax1 = fig.add_subplot(111)
    ax1.set_title(title + 'True vs Predicted')
    ax1.scatter(list(range(0,len(true))),true, s=10, c='r', marker="o", label='True')
    ax1.scatter(list(range(0,len(predicted))), predicted, s=10, c='b', marker="o", label='Predicted')
    plt.legend(loc='upper right');
    plt.show()    
    
##% Initial Models
SVM = svm.SVR().fit(train_X, train_y) 
RFReg = RandomForestRegressor(random_state = 0).fit(train_X, train_y)
XGReg = XGBRegressor(objective ='reg:squarederror', seed = 0,verbosity=0).fit(train_X,train_y) 
LGBMReg = LGBMRegressor(random_state=0).fit(train_X,train_y)
CatReg = CatBoostRegressor(random_seed=0).fit(train_X, train_y)

print("Support Vector Machine") 
predicted_train_y = SVM.predict(train_X)
evaluateRegressor(train_y,predicted_train_y,"    Training Set")
predicted_valid_y = SVM.predict(valid_X)
evaluateRegressor(valid_y,predicted_valid_y,"    Test Set")
print("\n")

##% Model Metrics
print("Random Forest Regressor") 
predicted_train_y = RFReg.predict(train_X)
evaluateRegressor(train_y,predicted_train_y,"    Training Set")
predicted_valid_y = RFReg.predict(valid_X)
evaluateRegressor(valid_y,predicted_valid_y,"    Test Set")
print("\n")
    
print("XGBoost Regressor") 
predicted_train_y = XGReg.predict(train_X)
evaluateRegressor(train_y,predicted_train_y,"    Training Set")
predicted_valid_y = XGReg.predict(valid_X)
evaluateRegressor(valid_y,predicted_valid_y,"    Test Set")
print("\n")

print("LightGBM Regressor") 
predicted_train_y = LGBMReg.predict(train_X)
evaluateRegressor(train_y,predicted_train_y,"    Training Set")
predicted_valid_y = LGBMReg.predict(valid_X)
evaluateRegressor(valid_y,predicted_valid_y,"    Test Set")

print("CatBoost Machine") 
predicted_train_y = CatReg.predict(train_X)
evaluateRegressor(train_y,predicted_train_y,"    Training Set")
predicted_valid_y = CatReg.predict(valid_X)
evaluateRegressor(valid_y,predicted_valid_y,"    Test Set")
print("\n")

# --------------------------------------------------------------------------------------------------
# Auto search for better hyper parameters with hyperopt, only need to give a range
# Reference: https://www.pythonf.cn/read/6998
#            https://lightgbm.readthedocs.io/en/latest/Parameters.html
#            https://lightgbm.readthedocs.io/en/latest/Parameters-Tuning.html#deal-with-over-fitting
#            https://lightgbm.readthedocs.io/en/latest/GPU-Performance.html
# 处理过拟合

#     设置较少的直方图数目 max_bin
#     设置较小的叶节点数 num_leaves
#     使用 min_child_samples（min_data_in_leaf） 和 min_child_weight（= min_sum_hessian_in_leaf）
#     通过设置 subsample（bagging_fraction） 和 subsample_freq（= bagging_freq） 来使用 bagging
#     通过设置 colsample_bytree（feature_fraction） 来使用特征子抽样
#     使用更大的训练数据
#     使用 reg_alpha（lambda_l1） , reg_lambda（lambda_l2） 和 min_split_gain（min_gain_to_split） 来使用正则
#     尝试 max_depth 来避免生成过深的树
#     Try extra_trees
#     Try increasing path_smooth

# trials = generate_trials_to_calculate([{'max_bin':63-8,               # default CPU 255 GPU 63
#                                         'max_depth':5-3,              # default -1
#                                         'num_leaves':31-20,           # default 31
#                                         'min_child_samples':20-10,    # default 20
#                                         'subsample_freq':1-1,         # default 1
#                                         'n_estimators':6000-1000,     # default 10
#                                         'learning_rate':0.01,         # default 0.1
#                                         'subsample':0.75,             # default 1.0
#                                         'colsample_bytree':0.8,       # default 1.0
#                                         'lambda_l1':0.0,              # default 0.0
#                                         'lambda_l2':0.0,              # default 0.0
#                                         'min_child_weight':0.001,     # default 0.001
#                                         'min_split_gain':0.0,         # default 0.0
#                                         #'path_smooth':0.0            # default 0.0
#                                         }])
# 缩小参数取值范围，搜索会快很多  
space_xgb = {
    'max_bin': hp.choice('max_bin', range(50, 501)),                  # CPU 50-501 GPU 8-128
    'max_depth': hp.choice('max_depth', range(3, 11)),    
    'n_estimators': hp.choice('n_estimators', range(100, 1001)),
    'learning_rate': hp.uniform('learning_rate', 0.01, 0.3),    
    'subsample': hp.uniform('subsample', 0.5, 0.99),
    'colsample_bytree': hp.uniform('colsample_bytree', 0.5, 0.99),
    'reg_alpha': hp.uniform('reg_alpha', 0, 5),                       # lambda_l1
    'reg_lambda': hp.uniform('reg_lambda', 0, 3),                     # lambda_l2
    'gamma': hp.uniform('gamma',0.0, 10),                             # min_split_loss, min_split_gain
    'min_child_weight': hp.uniform('min_child_weight',0.0001, 50),
}

def f_xgb(params):
    # Set extra_trees=True to avoid overfitting
    xgb = XGBRegressor(objective ='reg:squarederror', seed = 0,verbosity=0, **params)  # CPU 4.96s/trial
    #xgb = XGBRegressor(tree_method='gpu_hist', objective ='reg:squarederror', seed = 0,verbosity=0,**params)  # CPU 4.96s/trial    
    #xgb_model = xgb.fit(train_X, train_y)
    #acc = xgb_model.score(valid_X,valid_y)    
    # acc = cross_val_score(xgb, train_X, train_y).mean()              # CPU
    acc = cross_val_score(xgb, train_X, train_y, n_jobs=6).mean()  # GPU
    return -acc
# trials = Trials()
# Set initial values, start searching from the best point of GridSearchCV(), and default values
trials = generate_trials_to_calculate([{
                                        'max_bin':4,                                 # default 256
                                        'max_depth':5,                               # default 6
                                        'n_estimators':578,                          # default 100
                                        'learning_rate':0.05508679239402551,         # default 0.3
                                        'subsample':0.8429852720715357,              # default 1.0
                                        'colsample_bytree':0.8413894273173292,       # default 1.0
                                        'reg_alpha': 0.809791155072757,              # default 0.0
                                        'reg_lambda':1.4490119256389808,             # default 1.0
                                        'gamma':0.008478702584417519,                # default 0.0                                        
                                        'min_child_weight':24.524635200338793,       # default 1
                                        }])

t1 = time.time()  
# 1000trial [2:24:41,  8.68s/trial, best loss: -0.9080128034320879]                           
best_params = fmin(f_xgb, space_xgb, algo=tpe.suggest, max_evals=999, trials=trials)
t2 = time.time()
# 8681.310757875443
print("Time elapsed: ", t2-t1)

print('best:')
print(best_params)

# verify
params = best_params.copy()
# restore best hyper parameters
# {'colsample_bytree': 0.8413894273173292, 'gamma': 0.008478702584417519, 'learning_rate': 0.05508679239402551, 'max_bin': 4, 'max_depth': 5, 'min_child_weight': 24.524635200338793, 'n_estimators': 578, 'reg_alpha': 0.809791155072757, 'reg_lambda': 1.4490119256389808, 'subsample': 0.8429852720715357}
params =  {'colsample_bytree': 0.8413894273173292, 'gamma': 0.008478702584417519, 'learning_rate': 0.05508679239402551, 'max_bin': 4, 'max_depth': 5, 'min_child_weight': 24.524635200338793, 'n_estimators': 578, 'reg_alpha': 0.809791155072757, 'reg_lambda': 1.4490119256389808, 'subsample': 0.8429852720715357}
params['max_bin'] = params['max_bin']+50
params['max_depth'] = params['max_depth']+3
#params['num_leaves'] = params['num_leaves']+20
#params['min_child_samples'] = params['min_child_samples']+10
#params['subsample_freq'] = params['subsample_freq']+1
params['n_estimators'] = params['n_estimators']+100
print(params)

# Original best parameters of GridSearchCV()
# Set extra_trees=True to avoid overfitting
xgb_best = XGBRegressor(objective ='reg:squarederror', seed = 0,verbosity=0, **params)  # CPU 4.96s/trial

acc = cross_val_score(xgb_best, train_X, train_y).mean()
# 0.9080128034320879
print(acc)

# predict
xgb_model_full_data = xgb_best.fit(train_X, train_y)
# 0.9720045729712171
print(xgb_model_full_data.score(train_X,train_y))
# 0.9012739170573821
print(xgb_model_full_data.score(valid_X,valid_y))

# evalute model using the entire dataset from Train.csv
evaluateRegressor(train_y,xgb_model_full_data.predict(train_X),"Train set ")
evaluateRegressor(valid_y,xgb_model_full_data.predict(valid_X),"Valid set ")


def PlotPrediction2(true,predicted, title = "Dataset: "):
    df = pd.DataFrame({"Real":true,"Predicted":predicted})
    df.sort_values(by=["Real"], inplace=True)
    df = df.reset_index(drop=True)
    fig = plt.figure(figsize=(20,20))
    plt.tick_params(labelsize=20)
    ax1 = fig.add_subplot(111)
    ax1.set_title(title + 'True vs Predicted', fontsize=40)
    ax1.plot(list(range(0,len(df["Predicted"]))),df["Predicted"],'b-',label='Predicted')
    ax1.plot(list(range(0,len(df["Real"]))),df["Real"],'r-',label='True') 
    plt.xlabel("Samples in Value Order", fontsize=30)
    plt.ylabel("Log Price", fontsize=30)
    plt.legend(loc='upper left',fontsize=30 );
    plt.show()    
    
PlotPrediction2(train_y,xgb_model_full_data.predict(train_X),"Train set: ")
PlotPrediction2(valid_y,xgb_model_full_data.predict(valid_X),"Valid set: ")

##% Feature Importance 
df_importance = pd.DataFrame({"Feature":train_X.columns,"importance":xgb_model_full_data.feature_importances_})
df_importance.sort_values(["importance"],ascending = False, inplace=True)
plt.bar(df_importance["Feature"], df_importance["importance"])  
plt.xticks(rotation=45)
plt.show()

##% Feature Importance using shap package 
# https://zhuanlan.zhihu.com/p/83412330
# https://zhuanlan.zhihu.com/p/106320452
# Should use TreeExplainer for LightGBM tree base algorithms
shap_values = shap.TreeExplainer(xgb_best).shap_values(valid_X)
shap.summary_plot(shap_values, valid_X)
# a different importance plot of shap_values
shap.summary_plot(shap_values, valid_X, plot_type="bar")
# Feature importance for a single sample
# shap.initjs()  # notebook环境下，加载用于可视化的JS代码
# 如果不想用JS,传入matplotlib=True, 
row = 0
shap.force_plot(valid_y.mean(), shap_values[row], np.round(valid_X.iloc[row],4),matplotlib=True,text_rotation=30)
# better to plot with original feature values
shap.force_plot(valid_y.mean(), shap_values[row], valid_X2.iloc[row],matplotlib=True,text_rotation=30)
# matplotlib = True is not yet supported for force plots with multiple samples!
# shap.force_plot(valid_y.mean(), shap_values[0:3], X[0:3],matplotlib=True)
# create a SHAP dependence plot to show the effect of a single feature across the whole dataset
shap.dependence_plot("BuildingArea", shap_values, valid_X)

# print one trail to see the result structure
for t in trials:
    print(t)
    break

# hyperparameter is not relevant to a particular trial.
def unpack(x):
    if x:
        return x[0]
    return np.nan

# We'll first turn each trial into a series and then stack those series together as a dataframe.
trials_df = pd.DataFrame([pd.Series(t["misc"]["vals"]).apply(unpack) for t in trials])
# Then we'll add other relevant bits of information to the correct rows and perform a couple of
# mappings for convenience
trials_df["loss"] = [t["result"]["loss"] for t in trials]
trials_df["trial_number"] = trials_df.index
trials_df.to_csv("D:/temp/data/XGB_1000_trials.csv")
# 有一些迭代可能失败了，返回异常的数据。
trials_df = trials_df[trials_df['loss']<0]

def PlotTrial(trials,loss):
    fig = plt.figure(figsize=(20,20))
    plt.tick_params(labelsize=20)
    ax1 = fig.add_subplot(111)
    ax1.set_title('Loss per Trial', fontsize=40)
    ax1.plot(trials,loss,'b.',label='loss')
    # ax1.plot(list(range(0,len(df["Real"]))),df["Real"],'r-',label='True') 
    plt.xlabel("Trial", fontsize=30)
    plt.ylabel("Loss", fontsize=30)
    plt.legend(loc='upper left',fontsize=30 );
    plt.show()  
    
PlotTrial(trials_df['trial_number'], trials_df['loss'])    

trials_df["max_depth"].value_counts()
# draw contour plot with plotly in jupyter notebook then.



